from datetime import datetime
import base64
import json
import random
import string
import cherrypy  # also routes are needed
import cv2
import PIL
import hashlib
import jinja2
import tqdm
import pkgutil
from collections import OrderedDict
import zlib
import pathlib
import fargv
import sys
import os
import io
from PIL import Image
import cv2
from cherrypy.lib import file_generator
import numpy as np
import re


#grouting_js = pkgutil.get_data(__name__, "resources/grouting.js").decode("utf-8")
#grouting_jinja2 = pkgutil.get_data(__name__, "resources/grouting_webpage.jinja2").decode("utf-8")

frat_gui_js = pkgutil.get_data(__name__, "resources/frat_gui.js").decode("utf-8").encode('utf8')
frat_webpage_jinja2 = pkgutil.get_data(__name__, "resources/frat_webpage.jinja2").decode("utf-8")
empty_page_json = pkgutil.get_data(__name__, "resources/empty_page.json").decode("utf-8")
frat_gui_config = json.loads(pkgutil.get_data(__name__, "resources/font_annotation_config.json").decode("utf-8"))


def pillow_to_bytes(img, extention):
    img = np.array(img)
    if(len(img.shape)==3):
        img = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)
    status, res = cv2.imencode(f".{extention.lower()}", img)
    if status:
        return res.tobytes()
    else:
        raise ValueError

def create_thumb(fname, format="png", width=100, height=-1):
    img_bytes = open(fname, "rb").read()
    md5_str = hashlib.md5(img_bytes).hexdigest()
    img = Image.open(io.BytesIO(img_bytes))
    if height<0:
        old_width, old_height = img.size
        height = int((old_height/old_width)*width)
    img = img.resize((width, height))
    return pillow_to_bytes(img, format), md5_str

@cherrypy.expose
class FratWebServer(object):
    def load_from_local_images(self, image_filenames, img2itempath_regex, gt_filextention, autogt_filextention):
        self.thumbs = []
        self.image_names_to_idx = {}
        self.image_paths = []
        self.json_paths = []
        self.autojson_paths = []
        self.image_urls = []
        for n, image_filename in enumerate(tqdm.tqdm(image_filenames, desc="Creating thumbs")):
            thumb, md5id = create_thumb(image_filename)
            self.image_urls.append(f"{md5id}.png")
            self.thumbs.append(thumb)
            self.image_paths.append(image_filename)
            image_base_paths = re.findall(img2itempath_regex, image_filename)
            assert len(image_base_paths) >= 1 and len(image_base_paths[0]) > 0
            self.json_paths.append(image_base_paths[0]+gt_filextention)
            self.autojson_paths.append(image_base_paths[0]+autogt_filextention)
            self.image_names_to_idx[md5id] = n


    def __init__(self, image_filenames, image_web_format="png", html_template=None, config_dict=None) -> None:
        super().__init__()
        self.config_dict = {}
        if config_dict is None:
            self.config_dict.update(frat_gui_config)
        else:
            print(config_dict.keys())
            print()
            print(frat_gui_config.keys())
            print()
            print(set(config_dict.keys())-set(frat_gui_config))
            print()
            print(set(frat_gui_config)-set(config_dict.keys()))
            assert set(config_dict.keys()) == set(frat_gui_config.keys())
            self.config_dict.update(config_dict)

        if html_template == None:
            self.html_template = jinja2.Template(frat_webpage_jinja2)
        else:
            self.html_template = jinja2.Template(html_template)

        self.image_web_format = image_web_format
        print("INIT", self.config_dict)
        self.load_from_local_images(image_filenames, 
                                    img2itempath_regex=self.config_dict["img2itempath_regex"], 
                                    gt_filextention=self.config_dict["gt_filextention"],
                                    autogt_filextention=self.config_dict["autogt_filextention"])

        self.global_image_size_divisor = config_dict["image_size_divisor"]
        self.global_image_size_multiplier = config_dict["image_size_multiplier"]
        
    
    def scale_json(self, json_str):
        data = json.loads(json_str)
        new_data = data.copy()
        if "image_wh" in new_data.keys():
            new_data["image_wh"] = [(d*self.global_image_size_multiplier)//self.global_image_size_divisor for d in data["image_wh"]]
        new_data["rect_LTRB"] = []
        for rect in data["rect_LTRB"]:
            scaled_rect = [(d*self.global_image_size_multiplier)//self.global_image_size_divisor for d in rect]
            new_data["rect_LTRB"].append(scaled_rect)
        return json.dumps(new_data)


    def unscale_json(self, json_str):
        data = json.loads(json_str)
        new_data = data.copy()
        if "image_wh" in new_data.keys():
            new_data["image_wh"] = [(d*self.global_image_size_divisor)//self.global_image_size_multiplier for d in data["image_wh"]]
        new_data["rect_LTRB"] = []
        for rect in data["rect_LTRB"]:
            unscaled_rect = [(d*self.global_image_size_divisor)//self.global_image_size_multiplier for d in rect]
            new_data["rect_LTRB"].append(unscaled_rect)
        return json.dumps(new_data)



    def render_page_image(self, image_id):
        img_path = self.image_paths[self.image_names_to_idx[image_id]]
        img = Image.open(img_path)
        new_width = (self.global_image_size_multiplier*img.size[0])//self.global_image_size_divisor
        new_height = (self.global_image_size_multiplier*img.size[1])//self.global_image_size_divisor
        img = img.resize([new_width, new_height])
        cherrypy.response.headers['Content-Type'] = f"image/{self.image_web_format}"
        return pillow_to_bytes(img, self.image_web_format)
    
    def render_thumb(self, image_id):
        cherrypy.response.headers['Content-Type'] = f"image/{self.image_web_format}"
        return self.thumbs[self.image_names_to_idx[image_id]]
    
    def render_gt(self, image_id):
        cherrypy.response.headers['Content-Type'] = "application/json"
        json_path = self.json_paths[self.image_names_to_idx[image_id]]
        if(os.path.exists(json_path)):
            res_json = self.scale_json(open(json_path,"r").read())
        else:
            res_json = self.scale_json(empty_page_json)
        return res_json.encode('utf8')

    def render_html(self, image_id):
        cherrypy.response.headers['Content-Type'] = "text/html"
        return self.html_template.render(id=image_id, img_url=self.image_urls[self.image_names_to_idx[image_id]])

    def render_index(self):
        cherrypy.response.headers['Content-Type'] = "text/html"
        res=f"<html><body><table>"
        for n, name in sorted([(n, name) for name, n in self.image_names_to_idx.items()]):
            res+=f'<tr><td>{n}</td><td><a href="/{name}.html"><img src="/{name}.thumb.png"></a></td></tr>\n'
        res+="</table></body></html>"
        return res
    def render_idlist(self):
        cherrypy.response.headers['Content-Type'] = "application/json"
        all_ids = [v[1] for v in sorted([(v, k) for k, v  in self.image_names_to_idx.items()])]
        return str.encode(json.dumps(all_ids))


    @cherrypy.popargs('url_path')
    def GET(self, url_path=""):
        #print("GETTING:",repr(url_path))
        page_id = url_path.split("/")[-1].split(".")[0]
        if page_id == "page_id_list":
            return self.render_idlist()
        elif url_path.endswith(".thumb.png"):
            if page_id in self.image_names_to_idx:
                return self.render_thumb(page_id)
            else:
                raise cherrypy.HTTPError(404,"Thumb "+repr(page_id)+" not registered")
        elif url_path.endswith(".png"):
            if page_id in self.image_names_to_idx:
                return self.render_page_image(page_id)
            else:
                raise cherrypy.HTTPError(404,"Page "+repr(page_id)+" not registered. "+repr(self.image_names_to_idx))
        elif url_path.endswith(".html"):
            if page_id in self.image_names_to_idx:
                print(f"Serving {self.image_paths[self.image_names_to_idx[page_id]]}", file=sys.stderr)
                return self.render_html(page_id)
            else:
                raise cherrypy.HTTPError(404,"Groundtruth "+repr(page_id)+" not registered")
        elif url_path == "frat_config.json":
            cherrypy.response.headers['Content-Type'] = "application/javascript"
            return str.encode(json.dumps(self.config_dict))
        elif url_path.endswith(".json"):
            if page_id in self.image_names_to_idx:
                return self.render_gt(page_id)
            else:
                raise cherrypy.HTTPError(404,"Groundtruth "+repr(page_id)+" not registered")
        elif url_path == "favicon.ico":
            cherrypy.response.headers['Content-Type'] = "image/gif"
            return base64.decode("R0lGODlhAQABAAAAACH5BAEKAAEALAAAAAABAAEAAAICTAEAO")
        elif url_path == "frat_gui.js":
            cherrypy.response.headers['Content-Type'] = "application/javascript"
            return frat_gui_js
        elif url_path=="":
            return self.render_index()
        else:
            raise cherrypy.HTTPError(404,"Resource :"+repr(url_path)+"  id:"+repr(page_id)+" not understood")


    @cherrypy.tools.accept(media='application/json')
    def PUT(self, page_id):
        id_split = tuple(page_id.split("."))
        page_id = id_split[0]
        annotation_path = self.json_paths[self.image_names_to_idx[id_split[0]]]
        auto_annotation_path = self.autojson_paths[self.image_names_to_idx[id_split[0]]]
        #print("PUT:" + repr(page_id))
        cl = cherrypy.request.headers['Content-Length']
        json_string = str(cherrypy.request.body.read(int(cl)),'utf-8')
        json_string = self.unscale_json(json_string)


        img_md5 = page_id.split("/")[-1].split(".")[0]
        json_dict = json.loads(json_string)
        json_dict["img_md5"] = img_md5
        json_string = json.dumps(json_dict)

        if len(id_split) == 2 and id_split[1].lower()=="json": # normal annotation
            try:
                msg = f"{datetime.now().strftime('%m/%d/%Y, %H:%M:%S')}: saving ID{id_split[0]}"
                open(annotation_path, "w").write(json_string)
            except:
                raise
        elif len(id_split) == 3 and id_split[1:] == ("auto", "json"):
            if os.path.exists(annotation_path):
                annotation=open(annotation_path, "r").read()
            else:
                annotation=json_string
                open(annotation_path, "w").write(json_string)
            if os.path.exists(auto_annotation_path):
                auto_annotation=open(auto_annotation_path, "r").read()
            else:
                auto_annotation=annotation
            if auto_annotation!=json_string and annotation!=json_string:
                msg = f"{datetime.now().strftime('%m/%d/%Y, %H:%M:%S')}: autosaving {id_split[0]}"
                open(auto_annotation_path, "w").write(json_string)
            else:
                msg = f"{datetime.now().strftime('%m/%d/%Y, %H:%M:%S')}: not autosaving, all equal!"
        
        else:
            msg = f"{datetime.now().strftime('%m/%d/%Y, %H:%M:%S')}: PUT url:{page_id} could not be understood."
        print(msg)



from pathlib import Path

class GtImgAlligner:
    def __init__(self, img_ids) -> None:
        image_paths = []
        image_urls = []
        gt_paths = []
        gt_auto_paths = []
    
    def imgfname2jsonfname(img_id):
        pass
    
    def imgfname2jsonfname(img_id):
        pass

    def imgfname2jsonfname(img_id):
        pass
    



import numpy as np
from typing import Union, List, Tuple


def htmlrgb_to_uint8(color:Union[Tuple[str], List[str], str]):
    if isinstance(color, str):
        if "," in color:
            colors = tuple(color.split(","))
        else:
            colors = (color,)
    else:
        assert isinstance(color, list) or isinstance(color, tuple)
        colors = tuple(color)
    res = np.zeros([len(colors), 3], dtype=np.uint8)
    for n, color_str in enumerate(colors):
        color_str = color_str.lower()
        res[n, :] = [eval(f"0x{color_str[start: end]}") for start, end  in [(1,3), (3,5), (5,7)]]
    return res.tolist()


def rectangles_to_rgb(rect_list, rect_classes, output_size, color_palete, multiclass=False):
    if multiclass:
        return rectangles_to_rgb_multiclass(rect_list=rect_list, rect_classes=rect_classes,output_size=output_size, color_palete=color_palete)
    else:
        return rectangles_to_rgb_singleclass(rect_list=rect_list, rect_classes=rect_classes,output_size=output_size, color_palete=color_palete)
    

def rectangles_to_rgb_singleclass(rect_list, rect_classes, output_size, color_palete=[]):
    class_images = []
    output_size = output_size[1], output_size[0] # WxH to HxW
    integral_rgb = np.zeros(tuple(output_size)+(3,), dtype=np.int32)
    coverage_image = np.zeros([output_size[0], output_size[1]], dtype=np.int32)
    for classid in sorted(set(rect_classes)):
        #rgb = [classid%256, (classid//256)%256, (classid//65536)%256]
        #color_img = np.zeros([output_size[1], output_size[0]], 3, dtype=np.int32)
        rects = [rect for rect, rect_class in zip(rect_list,rect_classes) if classid==rect_class]
        for l,t,r,b in rects:
            print(f"l:{l}, t:{t}, r:{r}, b:{b} , img={output_size}")
            integral_rgb[t,l,:] += color_palete[classid]
            integral_rgb[b,r,:] += color_palete[classid]
            integral_rgb[t,r,:] -= color_palete[classid]
            integral_rgb[b,l,:] -= color_palete[classid]
            coverage_image[t,l] += 1
            coverage_image[b,r] += 1
            coverage_image[t,r] -= 1
            coverage_image[b,l] -= 1
    rgb = (integral_rgb.cumsum(axis=0).cumsum(axis=1)//coverage_image.cumsum(axis=0).cumsum(axis=1)[:,:,None])
    return rgb[:,:,[2,1,0]].astype(np.uint8)


def rectangles_to_rgb_multiclass(rect_list, rect_classes, output_size, color_palete):
    raise NotImplementedError
    class_images = []
    use_palete = color_palete is not None and len(color_palete) > 0 
    if use_palete:
        coverage_image = np.zeros([output_size[1], output_size[0]], dtype=np.int32)
    for classid in sorted(set(rect_classes)):
        #rgb = [classid%256, (classid//256)%256, (classid//65536)%256]
        if multiclass:
            class_val = 2**classid
        elif len(color_palete) > 0:
            rgb = color_palete[classid]
            class_val = eval(f"0x{rgb[1:]}")
        else:
            class_val = classid
        class_img = np.zeros([output_size[1], output_size[0]], dtype=np.int32)
        rects = [rect for rect, rect_class in zip(rect_list,rect_classes) if classid==rect_class]
        if len(rects):
            #print("I:", rgb, len(rects), "#{:06x}".format(class_val))
            print([class_val%256, (class_val//256)%256, (class_val//65536)%256])
        for l,t,r,b in rects:
            class_img[t:b,l:r]=class_val
        if use_palete:
            coverage_image += class_img==class_val
        class_images.append(class_img)
    class_images = np.stack(class_images)
    
    if multiclass and not use_palete:
        result = class_images.sum(axis=0)
        return np.stack([result%256, (result//256)%256, (result//65536)%256], axis=2).astype(np.uint8)
    if not multiclass and use_palete:
        result = class_images.sum(axis=0) // coverage_image
        rgb_result = np.stack([result%256, (result//256)%256, (result//65536)%256], axis=2) #// coverage_image[:,:,None]
        return rgb_result[:,:,::-1].astype(np.uint8)
    if not multiclass and not use_palete:
        result = class_images.max(axis=0)
        rgb_result = np.stack([result%256, (result//256)%256, (result//65536)%256], axis=2)
        return rgb_result.astype(np.uint8)
    else:
        raise ValueError("Impossible configuration")


def rectangles_to_gray(rect_list, rect_classes, output_size, multiclass=True):
    class_images = []
    for classid in sorted(set(rect_classes)):
        if multiclass:
            class_val = 2**classid
        else:
            class_val = classid
        class_img = np.zeros([output_size[1], output_size[0]], dtype=np.int32)
        rects = [rect for rect, rect_class in zip(rect_list,rect_classes) if classid==rect_class]
        for l,t,r,b in rects:
            class_img[t:b,l:r]=class_val
        class_images.append(class_img)
    class_images = np.stack(class_images)
    if multiclass:
        return class_images.sum(axis=0).astype(np.uint8)
    else:
        return class_images.max(axis=0).astype(np.uint8)



#from .web_server import StringGeneratorWebService, load_thumbd
from .generic_web_server import FratWebServer
from .version import __version__
import json

import pkgutil

frat_gui_js = pkgutil.get_data(__name__, "resources/frat_gui.js").decode("utf-8")
frat_gui_config_seals = json.loads(pkgutil.get_data(__name__, "resources/seals_annotation_config.json").decode("utf-8"))
frat_gui_config_font = json.loads(pkgutil.get_data(__name__, "resources/font_annotation_config.json").decode("utf-8"))
frat_gui_config = json.loads(pkgutil.get_data(__name__, "resources/font_annotation_config.json").decode("utf-8"))
frat_webpage_jinja2 = pkgutil.get_data(__name__, "resources/frat_webpage.jinja2").decode("utf-8")


from .formats import rectangles_to_rgb, rectangles_to_gray, htmlrgb_to_uint8


#import setuptools
from setuptools import setup, Extension
from distutils.util import convert_path

main_ns = {}
ver_path = convert_path('frat/version.py')
with open(ver_path) as ver_file:
    exec(ver_file.read(), main_ns)

setup(
    name='frat',
    version=main_ns['__version__'],
    packages=['frat'],
    #package_data={'frat': ['resources/diplomatics_annotation_config.json',
    #'frat/resources/font_annotation_config.json', 
    #'frat/resources/frat_webpage.jinja2',
    #'frat/resources/empty_page.json'
    #'frat/resources/frat_gui.js']},
    #include_package_data=True,
    package_data={'frat': ['resources/*.*']},
    scripts=['bin/frat', 'bin/frat_config', 'bin/frat_json2png', 'bin/frat_json2coco', 'bin/frat_json2yolov5'],
    license='GPLv3',
    author='Anguelos Nicolaou',
    author_email='anguelos.nicolaou@gmail.com',
    url='https://github.com/anguelos/frat',
    description="Fast Rectangle Annotation Tool",
    long_description_content_type="text/markdown",
    long_description=open('README.md').read(),
    #download_url='https://github.com/anguelos/tormentor/archive/0.1.0.tar.gz',
    keywords=["documents", "groudtruthing", "webbased"],
    classifiers=[
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering"],
    install_requires=["Pillow", "jinja2", "cherrypy", "fargv", "tqdm", "opencv-python-headless", "numpy", "opencv-python-headless"],
)